/*
 *  
 *  Copyright (c) 2004 Steve Slaven, All Rights Reserved.
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *  MA 02111-1307 USA
 *  
*/
package net.hoopajoo.cheeselab3;
import java.lang.reflect.Method;
import java.util.*;
import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.awt.event.*;

import javax.swing.*;
import javax.swing.Timer;

import net.hoopajoo.swingx.*;

// The main mapper panel
//class CheeseCanvas extends JPanel implements MouseListener,MouseMotionListener,KeyListener {
class CheeseCanvas extends JComponent implements CheeseMapListener,MouseListener,MouseMotionListener,KeyListener {
    final static int TOOL_SELECT = 0;
    final static int TOOL_MERGE = 1;
    final static int TOOL_MOVE = 2;
    final static int TOOL_DELETE = 3;
    final static int TOOL_POINT = 4;
    final String[] toolnames = { 
            "select_tool",
            "merge_tool",
            "move_tool",
            "delete_tool",
            "point_tool"
    };
    Cursor[] tool_cursors = {
            new Cursor( Cursor.DEFAULT_CURSOR ),
            new Cursor( Cursor.HAND_CURSOR ),
            new Cursor( Cursor.MOVE_CURSOR ),
            createCursor( "delete" ),
            new Cursor( Cursor.CROSSHAIR_CURSOR )
    };

    private int selectedid = 0;
    private int tool_active = TOOL_SELECT;
    private String tool_saved = "tool_select";
    private int ctx_edit_roomid = 0;
    private Point3d drag_start;
    private Point drag_start2d, drag_end2d;
    private boolean lclick_dragging = false;
    private boolean rclick_dragging = false;
    private PointToolCallback point_callback = null;
    private Cursor cursor_saved = getCursor();
    private Cursor cursor_movemap = new Cursor( Cursor.MOVE_CURSOR );
    private Cursor cursor_busy = new Cursor( Cursor.WAIT_CURSOR );

    // Gridded overlays, TODO: make this a gui option
    private boolean render_grid = true;

    // futures are buggy
    private boolean render_future = false;

    // Draw hud?
    private boolean render_hud = true;

    // TODO: These have no ui yet
    private boolean render_notes = false;
    private int render_notes_length = 15;

    // Will be turned on when refreshing (sync cache)
    protected boolean refreshing = false;

    // Selection handling
    private HashMap selected_rooms = new HashMap();
    private HashMap selected_areas = new HashMap();
    private Area selected_area = null;
    private float ants_phase = 0;
    
    // Graphical move handling
    private Integer[] moving_rooms = null;

    private int tile_width = 32, tile_height = 32,
            // Current origin info
            origin_x, origin_y, origin_z,

            // The origin coords at time of last map rendering
            // saved so that players and stuff overlay properly
            // when the map hasn't refreshed yet, otherwise they
            // will jump around the map during heavy movement
            map_origin_x, map_origin_y, map_origin_z,
            remap_origin_x, remap_origin_y, remap_origin_z;

    private String theme;
    protected long changed = 0,
    // this basically tracks if we need to redraw the mapcache (new rooms, map
    // data, resize, etc)
            changed_map = 0,
            last_map = 0,
            last_sync = 0,
            last_mapdata = -1;

    /* We use the VolatileImage stuff for OSC-s.
     * The acceleration does not work very well under Linux yet, but sun
     * is working on it and we want to be ready when it hits the streets.
     * Also under Windows with Java 1.5 (and under Linux with Java 1.6 beta)
     * we can get that fuzzy feeling which comes from seeing that some of our
     * tiles are indeed marked as "accelerated" (Help->System Information).
     * 
     * However, this VolatileImage stuff is not yet correctly implemented
     * so we stick with BufferedImages for now.  
     */
    private VolatileImage screencache = null;
    private VolatileImage mapcache = null;
    //private BufferedImage screencache = null;
    //private BufferedImage mapcache = null;

    protected CheeseRenderer renderer;

    // TODO: Needs another map for temp overlays, that can be image/text
    // so that we can "blink" people off where. These overlays
    // should have an expire time (+5 seconds?) that will make them
    // auto-delete at the next render after the expiry time
    // this would also make the "hud" display of hp/move/mana possible

    // This is here so that the tilenames are standard for all renderers
    private HashMap tilenames = new HashMap();

    // Event handlers
    private Vector listeners = new Vector();

    protected String follow_userid = "";
    protected boolean follow_player = true;

    public SelectedAction follow_action;
    public SelectedAction zoomin_action;
    public SelectedAction zoomout_action;
    public SelectedAction levelup_action;
    public SelectedAction leveldown_action;

    CheeseCanvas() {
        addMouseListener( this );
        addMouseMotionListener( this );
        addKeyListener( this );
        setToolTipText( "Room Info" );

        // Set tile mapping, this provides common tile names for all renderers
        tilenames.put("@","indoors");
        tilenames.put("<","mountain");
        tilenames.put("+","road");
        tilenames.put("#","city");
        tilenames.put("~","shallow_water");
        tilenames.put("\"","brush");
        tilenames.put("_","field");
        tilenames.put("w","water");
        tilenames.put("=","tunnel");
        tilenames.put("%","forest");
        tilenames.put("W","rapids");
        tilenames.put("O","cavern");
        tilenames.put("(","hills");
        tilenames.put("U","underwater");

        String r = CheeseMap.prop_get_string( "renderer" );
        if( r.equals( "cached" ) ) {
            CheeseLab.dmesg( 1, "Enabling experimental cached renderer" );
            renderer = new CheeseRendererCached();
        }else if( r.equals( "basic" ) ) {
            CheeseLab.dmesg( 1, "Enabling basic renderer" );
            renderer = new CheeseRendererBasic();
        }else if( r.equals( "threaded" ) ) {
            CheeseLab.dmesg( 1, "Enabling experimental threaded renderer" );
            renderer = new CheeseRendererThreaded();
        }else{
            // Default for wacky values
            CheeseLab.dmesg( 1, "Unknown renderer: '" + r + "', using basic renderer" );
            renderer = new CheeseRendererBasic();
        }
        renderer.set_canvas( this );

        /* This must be done after instantiating the renderer because othervise if the
         * renderer also registers itself as a CheeseMapListener we get the events before
         * the renderer does and when we call for redraw, the renderers internal caches
         * will be out of date
         */
        CheeseMap.addCheeseMapListener( this );

        /* Create action objects for controlling zoom,
         * follow player etc from different GUI controls */
        follow_action = new SelectedAction() {
                            public void actionPerformed(ActionEvent e) {
                                set_follow_player( !isSelected() );
                            }
                        };
        follow_action.putValue( SelectedAction.SHORT_DESCRIPTION, "Follow Player" );
        follow_action.putValue( SelectedAction.ACCELERATOR_KEY,
                                KeyStroke.getKeyStroke( 'f' ) );

        zoomin_action = new SelectedAction() {
                            public void actionPerformed( ActionEvent e ) {
                                zoom_in();
                            }
                        };
        zoomin_action.putValue( SelectedAction.SHORT_DESCRIPTION, "Zoom IN (+)" );
        zoomin_action.putValue( SelectedAction.ACCELERATOR_KEY,
                                KeyStroke.getKeyStroke( '+' ) );

        zoomout_action = new SelectedAction() {
            public void actionPerformed(ActionEvent e) {
                zoom_out();
            }
        };
        zoomout_action.putValue( SelectedAction.SHORT_DESCRIPTION, "Zoom OUT (-)" );
        zoomout_action.putValue( SelectedAction.ACCELERATOR_KEY,
                                 KeyStroke.getKeyStroke( '-' ) );

        levelup_action = new SelectedAction() {
            public void actionPerformed(ActionEvent e) {
                level_up();
            }
        };
        levelup_action.putValue( SelectedAction.SHORT_DESCRIPTION, "UP one level" );
        /* Use page_up now (not bound to this event either)
        levelup_action.putValue( SelectedAction.ACCELERATOR_KEY,
                KeyStroke.getKeyStroke( KeyEvent.VK_UP,
                                        InputEvent.CTRL_MASK ) );
        */

        leveldown_action = new SelectedAction() {
            public void actionPerformed(ActionEvent e) {
                level_down();
            }
        };
        leveldown_action.putValue( SelectedAction.SHORT_DESCRIPTION, "DOWN one level" );
        /* Use page_down now (not bound to this event)
        leveldown_action.putValue( SelectedAction.ACCELERATOR_KEY,
                KeyStroke.getKeyStroke( KeyEvent.VK_DOWN,
                                        InputEvent.CTRL_MASK ) );
        */

        // Create a timer for the "marching ants" effect
        Timer t = new Timer( 250, new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                // Request repaint if there is an active selection
                if( selected_area != null ) {
                    changed();
                }
            }
        } );
        t.start();
        
        setBackground( Color.black );
        setDoubleBuffered( false );
        setFocusable( true );
        // requestDefaultFocus();
        // requestFocus();
        requestFocusInWindow();
    }


    public void mouseMoved(MouseEvent e) {
    }

    public void mouseDragged(MouseEvent e) {
        if( lclick_dragging ) {
            // draggin to select
            drag_end2d = e.getPoint();
            repaint();
        }
        if( rclick_dragging ) {
            // dragging the map around
            drag_end2d = e.getPoint();
            repaint();
        }
    }

    public void mousePressed(MouseEvent e) {
        java.awt.Point p = e.getPoint();

        if( e.getButton() == 1 ) {
            CheeseLab.dmesg( 3, "Button 1 down" );

            drag_start = local_to_global(p.x, p.y);
            drag_start2d = p;
            drag_end2d = p; // ends here too atm
            lclick_dragging = true;
            
            // Prepare room move rendering if move tool is selected
            if( tool_active == TOOL_MOVE ) {
                boolean rv;
                
                if( selected_rooms.isEmpty() ) {
                    // Get the room we were clicked on
                    CheeseRoomData room = CheeseMap.map.get_room( drag_start.x,
                            drag_start.y, drag_start.z );
                    if( room != null )
                        moving_rooms = new Integer [] { new Integer( room.roomid ) };
                }else{
                    moving_rooms = (Integer [])selected_rooms.keySet().
                                                    toArray( new Integer[0] );
                }
                rv = renderer.prepare_move( moving_rooms );

            }

        }else if( e.getButton() == 3 ) {
            CheeseLab.dmesg( 3, "Button 3 down" );
            drag_start = local_to_global( p.x, p.y );
            drag_start2d = p;
            drag_end2d = p; // ends here too atm
            rclick_dragging = true;
            cursor_saved = getCursor();
            setCursor( cursor_movemap );
        }
    }

    public void mouseReleased(MouseEvent e) {
        java.awt.Point p = e.getPoint();
        Point3d drag_end = local_to_global( p.x, p.y );

        if( e.getButton() == 1 ) {
            if( tool_active == TOOL_SELECT ) {
                boolean select_3d = false;

                if(!e.isShiftDown()) {	// Allow adding rooms to selection by holding the SHIFT key
                    selected_rooms.clear();
                    selected_areas.clear();
                }

                if(e.isControlDown())	// Do 3D select if Control was held down
                    select_3d = true;

                CheeseLab.dmesg( 3, "Button 1 up" );

                int x, y, z,
                min_x, max_x,
                min_y, max_y,
                min_z, max_z;

                if( select_3d ) {
                    min_z = CheeseMap.map.stats.min_z;
                    max_z = CheeseMap.map.stats.max_z;
                } else {
                    min_z = max_z = drag_start.z;
                }

                if( drag_start.x < drag_end.x ) {
                    min_x = drag_start.x;
                    max_x = drag_end.x;
                }else{
                    min_x = drag_end.x;
                    max_x = drag_start.x;
                }

                if( drag_start.y < drag_end.y ) {
                    min_y = drag_start.y;
                    max_y = drag_end.y;
                }else{
                    min_y = drag_end.y;
                    max_y = drag_start.y;
                }

                int nareas = 0;
                for( z = min_z; z <= max_z; z++ ) {
                    CheeseLab.dmesg( 3, "Selecting rooms from: " + min_x + "," + min_y + "," + z +
                                     " " + max_x + "," + max_y + "," + z );
                    
                    Area a = (Area)selected_areas.get( new Integer( z ) );
                    Rectangle r = new Rectangle( min_x, min_y, 0, 1 );

                    for( y = min_y; y <= max_y; y++ ) {
                        r.x = min_x; r.y = y; r.width = 0;
                        for( x = min_x; x <= max_x; x++ ) {
                            // These rooms ought to be in our map since you can't really
                            // click on them unless you can see them
                            CheeseRoomData room = CheeseMap.map.get_room( x, y, z );
                            if( room != null ) {
                                Integer roomid = new Integer( room.roomid );
                                CheeseLab.dmesg( 3, "Selecting: " + room.roomid );
                                selected_rooms.put( roomid, roomid );
                                r.width++;
                            }

                            /* We have to optimize area creation a bit, since 
                             * creating a new Area object and adding it to an existing
                             * one is horribly slow for some unknow reason.
                             * Thus we try to reduce the number of additions by
                             * extending a rectangle, 1 room high to as far right as
                             * possible. 
                             */
                            
                            // In case of broken continuity or last x value
                            if( room == null || x == max_x ) {
                                if( r.width > 0 ) {
                                    // Add to selection area
                                    if( a == null ) {
                                        a = new Area( r );
                                        selected_areas.put( new Integer( z ), a );
                                    }else{
                                        a.add( new Area( r ) );
                                    }
                                    nareas++;
                                }
                                r.x = x + 1; r.width = 0;
                            }
                        }
                    }
                }
                
                CheeseLab.dmesg( 1, "Added " + selected_rooms.size() + " rooms to selection" );
                CheeseLab.dmesg( 1, "Created " + nareas + " Area objects in the process" );

                /* TODO: Maybe we should move map if there was only one room selected.
                 *       The need for LMB navigation is debatable though
                 * 
                 * if( selected_rooms.size() == 1 ) {
                 *	Point3d p3d = local_to_global( p.x, p.y );
                 *	center_on( p3d.x, p3d.y, p3d.z );
                 *	CheeseMap.set_follow_player( false );
                 * }
                 */

                updateSelectionArea();

                requestFocus();
            }else if( tool_active == TOOL_MOVE ) {
                // graphical move
                // FIXME: implement single room under cursor move
                int dx = drag_end.x - drag_start.x;
                int dy = drag_end.y - drag_start.y;
                
                /* If there was any rooms that were moved and they did
                 * end up in a different location after all */
                if( moving_rooms != null && ! (dx == 0 && dy == 0 ) ) {
                    CheeseMap.move_rooms( moving_rooms, dx, dy, 0 );
                }
            }else if( tool_active == TOOL_DELETE ) {
                // delete
                CheeseRoomData room = CheeseMap.map.get_room( drag_end.x, drag_end.y, drag_end.z );
                if( room != null ) {
                    CheeseMap.delete_room( room.roomid );
                }
            }else if( tool_active == TOOL_MERGE ) {
                // merge
                CheeseRoomData room = CheeseMap.map.get_room( drag_end.x, drag_end.y, drag_end.z );
                if( room != null ) {
                    CheeseMap.send_tx( 1, "merge", room.roomid + "" );
                }
            }else if( tool_active == TOOL_POINT ) {
                // Restore last tool. Execute callback.
                setEditTool( tool_saved );
                if (point_callback != null) {
                    point_callback.pointedTo(p);
                    point_callback = null;
                }
            }

            lclick_dragging = false;
            /*
            remap(); // get new view data in case rooms moved
            */
            // map_changed(); // force redraw of cached map
            changed(); // call repaint to update display
        }

        if( e.getButton() == 3 ) {
            int dx = drag_start.x - drag_end.x;
            int dy = drag_start.y - drag_end.y;

            rclick_dragging = false;
            setCursor( cursor_saved );

            if( dx != 0 || dy != 0 ) {
                move_origin( dx, dy, 0 );
                set_follow_player( false );
            }else{
                // Call repaint to cancel visual effect of "accidental" dragging
                repaint();

                // Now build and open context menu
                JPopupMenu menu = new JPopupMenu();
                JMenuItem item;
                CheeseRoomData room = null;

                // Context menu room editor
                String r_name;
                room = CheeseMap.map.get_room( drag_end.x, drag_end.y, drag_end.z );
                if( room != null ) {
                    r_name = room.name;
                    ctx_edit_roomid = room.roomid;
                }else{
                    r_name = "No Room Here";
                    ctx_edit_roomid = 0;
                }

                // Trim room name length to show it in context menu
                if( r_name.length() > 25 )
                    r_name = r_name.substring( 0, 22 ) + "... ";

                item = new JMenuItem( r_name );
                item.setEnabled( false );
                menu.add( item );
                menu.addSeparator();

                if( ctx_edit_roomid != 0 ) {
                    item = new JMenuItem( "Open Room Editor" );
                    item.addActionListener( new ActionListener() {
                        public void actionPerformed(ActionEvent e) {
                            selectedid = ctx_edit_roomid;
                            CheeseLab.dmesg( 3, "Do editor for: "
                                    + ctx_edit_roomid );
                            CheeseMap.set_edit_room( CheeseMap.map
                                    .get_room( ctx_edit_roomid ) );
                        }
                    } );
                    menu.add(item);

                    // Construct sub-menu for Overlay editing
                    String [] ovrl = CheeseMapRoomEditor.common_overlays;
                    JMenu submenu = new JMenu( "Overlays" );
                    for( int i = 0; i < ovrl.length; i++) {
                        boolean checked = false;
                        ImageIcon icon = new ImageIcon(
                                             get_image( "overlay_" + ovrl[i], 16, 16 ) );
                        for( int j = 0; j < room.overlays.length; j++ ) {
                            if( room.overlays[j].equals( ovrl[i] ) )
                                checked = true;
                        }
                        item = new JCheckBoxMenuItem( ovrl[i], icon, checked );
                        item.addActionListener( new CtxEditorListener(
                                                    ovrl[i], ctx_edit_roomid ) );
                        item.setActionCommand( "set_overlay" );
                        submenu.add( item );
                    }
                    menu.add( submenu );

                    // Construct sub-menus for exit editing
                    String[] dirs = { "N", "S", "E", "W", "U", "D" };
                    int[] cur_exits = { room.exit_n, room.exit_s, room.exit_e,
                                        room.exit_w, room.exit_u, room.exit_d };

                    for( int i = 0; i < dirs.length; i++ ) {
                        JMenu m = new JMenu( "Exit " + dirs[i] );
                        ButtonGroup bg = new ButtonGroup();

                        bg.add( item = new JRadioButtonMenuItem( "None" ) );
                        if( cur_exits[i] == 0 )
                            item.setSelected( true );
                        item.addActionListener( new CtxEditorListener( dirs[i], ctx_edit_roomid ) );
                        item.setActionCommand( "set_exit_none" );
                        m.add( item );
                        bg.add( item = new JRadioButtonMenuItem( "Unexplored" ) );
                        if( cur_exits[i] == -1 )
                            item.setSelected( true );
                        item.addActionListener( new CtxEditorListener( dirs[i], ctx_edit_roomid ) );
                        item.setActionCommand( "set_exit_unexplored" );
                        m.add( item );
                        bg.add( item = new JRadioButtonMenuItem( "Death Trap" ) );
                        if( cur_exits[i] == -2 )
                            item.setSelected( true );
                        item.addActionListener( new CtxEditorListener( dirs[i], ctx_edit_roomid ) );
                        item.setActionCommand( "set_exit_dt" );
                        m.add( item );
                        bg.add( item = new JRadioButtonMenuItem( "To room ..." ) );
                        if( cur_exits[i] > 0 ) {
                            item.setText( "To room " + cur_exits[i] );
                            item.setSelected( true );
                        }
                        item.addActionListener( new CtxEditorListener( dirs[i], ctx_edit_roomid ) );
                        item.setActionCommand( "set_exit_from_map" );
                        m.add( item );

                        menu.add( m );
                    }
                }

                // Main operations
                menu.addSeparator();
                menu.add( new JSelectedCheckMenuItem( follow_action ) );
                menu.add( new JSelectedMenuItem( levelup_action ) );
                menu.add( new JSelectedMenuItem( leveldown_action ) );
                menu.add( new JSelectedMenuItem( zoomin_action ) );
                menu.add( new JSelectedMenuItem( zoomout_action ) );

                item = new JMenuItem( "Merge Selected" );
                item.addActionListener( new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        merge_selected();
                    }
                } );
                menu.add(item);

                // Toolbox
                menu.add( CheeseMap.toolbox.createMenu( "Tools" ) );

                // Phew! Done! Show it.
                // menu.setLightWeightPopupEnabled(false);
                menu.show( this, p.x, p.y );
            }
        }
    }

    public void mouseEntered(MouseEvent e) { }
    public void mouseExited(MouseEvent e) { }
    public void mouseClicked(MouseEvent e) { }

    public String getToolTipText( MouseEvent e ) {
        //if( ! online )
        //    return null;

        java.awt.Point p = e.getPoint();

        Point3d p3d = local_to_global( p.x, p.y );

        // TODO: make this useful :p
        String ret = new String();
        CheeseRoomData room = CheeseMap.map.get_room( p3d.x, p3d.y, p3d.z );
        if( room != null ) {
            String show_notes = room.notes;
            if( show_notes == null )
                show_notes = "";

            if( ! show_notes.equals( "" ) )
                show_notes = "<br>" + show_notes;

            // Check for hidden exits
            String show_hidden = "";
            if( ( ! room.door_n.equals( "" ) ) && ( ! room.door_n.equals( "exit" ) ) )
                show_hidden += "<br>Door north: " + room.door_n;

            if( ( ! room.door_s.equals( "" ) ) && ( ! room.door_s.equals( "exit" ) ) )
                show_hidden += "<br>Door south: " + room.door_s;

            if( ( ! room.door_e.equals( "" ) ) && ( ! room.door_e.equals( "exit" ) ) )
                show_hidden += "<br>Door east: " + room.door_e;

            if( ( ! room.door_w.equals( "" ) ) && ( ! room.door_w.equals( "exit" ) ) )
                show_hidden += "<br>Door west: " + room.door_w;

            if( ( ! room.door_u.equals( "" ) ) && ( ! room.door_u.equals( "exit" ) ) )
                show_hidden += "<br>Door up: " + room.door_u;

            if( ( ! room.door_d.equals( "" ) ) && ( ! room.door_d.equals( "exit" ) ) )
                show_hidden += "<br>Door down: " + room.door_d;

            ret = "<html>" +
                  "#" + room.roomid + " - " + room.x + "," + room.y + "," + room.z + "<br>" +
                  room.name +
                  show_notes +
                  show_hidden +
                  "</html>";
        }else{
            //ret = "No Room Here";
            ret = null;
        }

        return( ret );
    }

    public void paintComponent(Graphics g) {

        if( CheeseMap.online == false ) {
            super.paintComponent( g );
            return;
        }
        
        // Here we handle all this outrageous VolatileImage stuff
        if( screencache == null ) newScreenCache();
        if( mapcache    == null ) newMapCache();
        
        GraphicsConfiguration gc = getGraphicsConfiguration();

        do {
            int vc1 = screencache.validate( gc );
            int vc2 = mapcache.validate( gc );
            if( vc1 == VolatileImage.IMAGE_RESTORED ) {
                CheeseLab.dmesg( 2, "Surface loss on OSC detected." );
                changed( false );
            }else if( vc1 == VolatileImage.IMAGE_INCOMPATIBLE ) {
                CheeseLab.dmesg( 2, "OSC VolatileImage to be recreated." );
                newScreenCache();
                changed( false );
            }
            if (vc2 == VolatileImage.IMAGE_RESTORED) {
                CheeseLab.dmesg( 2, "Surface loss on Map OSC detected." );
                map_changed();
                changed( false );
            }else if (vc2 == VolatileImage.IMAGE_INCOMPATIBLE) {
                CheeseLab.dmesg( 2, "Map OSC VolatileImage to be recreated." );
                newMapCache();
                map_changed();
                changed( false );
            }
            renderComponent( g );
        } while ( screencache.contentsLost() || mapcache.contentsLost() );

    }
    
    private void renderComponent(Graphics g) {
        // Handle right-click map move mode
        if( rclick_dragging ) {
            if( screencache != null ) // We can't move around nonexisting picture right?
                g.drawImage( screencache, drag_end2d.x - drag_start2d.x,
                             drag_end2d.y - drag_start2d.y, null );
            return;
        }
        
        // Handle resize refreshes
        if( screencache != null ) {
            // validate width/height
            if( screencache.getWidth(null) != getWidth() ) {
                screencache.flush();
                screencache = null;
            }else if( screencache.getHeight(null) != getHeight() ) {
                screencache.flush();
                screencache = null;
            }

            // Send view changed events on resize
            if( screencache == null ) {
                sendEvents();

                // take player position into account if in follow_player mode
                set_follow_player( follow_player );
            }
        }
        
        // Make our double buffer
        if( screencache == null ) {
            newScreenCache();
            remap(); // get updated view since view changed
            changed( false );
        }
        
        /* Handle move tool */
        if( lclick_dragging && tool_active == TOOL_MOVE ) {
            // Calculate how far to move the rooms
            Point3d drag_end = local_to_global( drag_end2d.x, drag_end2d.y );
            
            renderer.draw_move( screencache,
                    drag_end.x - drag_start.x, drag_end.y - drag_start.y );
        } else 

        // Paint to a bitmap, that way exposed paint events don't flicker like mad
        if( changed != last_sync || screencache == null ) {
            CheeseLab.dmesg( 2, "Something changed: c" +
                             changed + "/" + last_sync + " m" +
                             changed_map + "/" + last_map  );

            // FIXME: Race condition! changed() can be called from other thread.
            last_sync = changed;

            // Make sure the map matches the size of osc
            if( mapcache != null ) {
                if( mapcache.getWidth() != screencache.getWidth() ) {
                    mapcache.flush();
                    mapcache = null;
                }else if( mapcache.getHeight() != screencache.getHeight() ) {
                    mapcache.flush();
                    mapcache = null;
                }
            }

            // Create the mapcache if needed
            if( mapcache == null ) {
                newMapCache();
                map_changed();
            }

            if( refreshing ) {
                // Do not draw map if we are in the middle of a refresh
                // we will get our chance with the refresh finished event
            }
            // Update base map if we got new map data
            else if( last_map < changed_map ) {
                /* FIXME: Race condition!
                 * This line must be as close as possible to the if that
                 * checks those variables. Failing to do that will lead to
                 * unsolvable race conditions, because map_changed can be
                 * called from other threads. It would be real nice if we 
                 * could check and increment as a same atomic operation, but
                 * this would require locks between this part here and
                 * map_changed() which we don't feel like adding.
                 */
                last_map = changed_map;

                // Request updated info (happens in the background)
                if( remap_origin_x != origin_x ||
                        remap_origin_y != origin_y ||
                        remap_origin_z != origin_z ) {
                    // Any time our origin changes we explicitly call remap
                    // here to make sure we are up-to-date
                    remap();
                }else{
                    // This should work because we move the origin, get a redraw,
                    // which causes a remap (above) then render our cached view.
                    // when the map command finishes, it also requests a repaint,
                    // but the origin doesn't move so we end up here (remap==origin)
                    // and call the catchup.  So that way the final remap will always
                    // trigger a catchup remap if we don't move and make sure our
                    // *current* view is updated
                    CheeseMap.map.catchup_remap();
                }

                // Set all those nice map_origin_ variables to point to current origin
                updateMapOrigin();
                // Let the code bootstrap
                renderer.prepare();

                Date rstart = new Date();

                renderer.draw_map( mapcache );

                Date rend = new Date();
                long rdiff = rend.getTime() - rstart.getTime();
                double secs = (double)rdiff / 1000;
                CheeseLab.dmesg( 2, "Render time: " + rdiff + "ms (" + secs + " secs)" );
            }else{
                // Initialize this render frame, no origins need to change
                renderer.prepare();
            }

            // Now render map data + players + hud to screencache
            Graphics2D sg = (Graphics2D)screencache.getGraphics();
            sg.drawImage( mapcache, 0, 0, null );
            

            renderer.draw_players( screencache );

            if( render_hud )
                renderer.draw_hud( screencache );

            // Draw selection outline
            if( selected_area != null ) {
                float line = tile_width / 10f;
                ants_phase += line;
                float dash1[] = { 5 * line };
                BasicStroke dashed = new BasicStroke( line, BasicStroke.CAP_BUTT,
                        BasicStroke.JOIN_MITER, 5.0f, dash1, ants_phase );
                
                sg.setColor( Color.white );
                sg.setStroke( dashed );
                sg.draw( selected_area );
            }

            sg.dispose();
        }
        
        g.drawImage( screencache, 0, 0, null );

        // if dragging draw that
        if( lclick_dragging && tool_active == TOOL_SELECT ) {
            int rsx, rsy, rex, rey;

            CheeseLab.dmesg( 3, "Drawing rect" );
            Point3d drag_end = local_to_global( drag_end2d.x, drag_end2d.y );
            
            Point s = global_to_local( drag_start.x, drag_start.y, drag_start.z );
            Point e = global_to_local( drag_end.x, drag_end.y, drag_end.z );
            
            if( s.x > e.x ) { rsx = e.x; rex = s.x - e.x + tile_width; }
            else            { rsx = s.x; rex = e.x - s.x + tile_width; }
            if( s.y > e.y ) { rsy = e.y; rey = s.y - e.y + tile_height; }
            else            { rsy = s.y; rey = e.y - s.y + tile_height; }

            g.setColor( Color.RED );
            g.drawRect( rsx, rsy, rex, rey);
        }
    }

    /**
     * Re-calculate area object used to draw selection outline (marching ants) to screen
     *
     */
    private void updateSelectionArea() {
        Area a = (Area)selected_areas.get( new Integer( origin_z ) );
        selected_area = a;
        
        if( a != null ) {
            AffineTransform trans = new AffineTransform();
            trans.scale( tile_width, tile_height );
            trans.translate( -origin_x, -origin_y );
            selected_area = a.createTransformedArea( trans );
        }
    }

    public Point global_to_local_reference( int x, int y, int z, int rx, int ry, int rxz ) {
        Point p = new Point();
        p.x = ( x - rx ) * tile_width;
        //p.y = ( rows - ( y - origin_y ) ) * tile_height;
        p.y = ( y - ry ) * tile_height;
        return( p );
    }

    public Point global_to_local( int x, int y, int z ) {
        return( global_to_local_reference( x, y, z, origin_x, origin_y, origin_z ) );
    }

    public Point3d local_to_global( int x, int y ) {
        Point3d p = new Point3d();

        p.x = origin_x + ( x / tile_width );
        //p.y = origin_y + ( rows - ( y / tile_width ) );
        p.y = origin_y + ( y / tile_height );
        p.z = origin_z;

        return( p );
    }

    public void center_on( int x, int y, int z ) {
        CheeseLab.dmesg( 2, "Centering on: " + x + "," + y + "," + z );
        int cols = getWidth() / tile_width;
        int rows = getHeight() / tile_height;

        set_origin( x - ( cols / 2 ), y - ( rows / 2 ), z );
    }

    public void move_origin( int dx, int dy, int dz ) {
        CheeseLab.dmesg( 2, "Moving: " + dx + "," + dy + "," + dz );
        set_origin( origin_x + dx, origin_y + dy, origin_z + dz );
    }

    public void set_origin( int x, int y, int z ) {
        CheeseLab.dmesg( 2, "Setting origin: " + x + "," + y + "," + z );

        // No need to refresh anything. if origin really didn't move
        if( origin_x == x && origin_y == y && origin_z == z )
            return;

        origin_x = x;
        origin_y = y;
        origin_z = z;

        // Not necessary, the draw code will do this if needed
        //remap();

        sendEvents();

        map_changed();
        changed();
    }

    // Request updated map data, this should be called whenever the
    // origin moves or if rooms are moved around (world change events)
    // This will in-advertantly trigger a repaint
    public void remap() {
        // We mark the sync now so that if there are changed while we
        // are rendering, we will get them
        int rend_x = origin_x,
            rend_y = origin_y,
            rend_z = origin_z;

        remap_origin_x = rend_x;
        remap_origin_y = rend_y;
        remap_origin_z = rend_z;

        int cols = getWidth() / tile_width;
        int rows = getHeight() / tile_height;

        CheeseLab.dmesg( 2, "Remap" );
        CheeseLab.dmesg( 3, "Size: " + rows + "," + cols );
        CheeseLab.dmesg( 3, "Origin: " + rend_x + "," + rend_y + "," + rend_z );

        // Figure our bounding box
        int start_x = rend_x;
        int start_y = rend_y;
        int start_z = rend_z;

        int end_x = start_x + cols;
        int end_y = start_y + rows;
        int end_z = rend_z + 1;

        long tmp_map = changed_map;

        // Request this area of the map to be updated, it will repaint if
        // needed (on room updates)
        CheeseMap.map.remap( start_x, start_y, start_z, end_x, end_y, end_z );

        /* Wait for mapdata to be updated, but not for longer than 20ms.
         * This will save us a repaint in most cases and users will not
         * notice a map lagging less than 20ms behind.
         * FIXME: Actually we shouldn't draw the map with only partial mapdata
         *        available. This will have an effect that unloaded rooms will
         *        initially not appear on the map at all (they should be marked
         *        with ??-s) and the respective exits will be drawn as warps.
         */
        /* TODO: This does not work well with area caching, there is no way to
         *       notify canvas from CheeseMapData, however the server trip will
         *       almost always take more than 20ms according to my tests :(
        final int max_wait_remap = 20;
        int wait = 0;
        for( wait = 0; wait < max_wait_remap && tmp_map == changed_map; wait++ ) {
            try {
                Thread.sleep( 1 );
            }catch(InterruptedException e) {}
        }
        if( wait > 0 )
            CheeseLab.dmesg(2, "Waited " + (( wait < max_wait_remap ) ? "" : ">") + 
                    wait + "ms for map data from server.");
        */
    }

    public void map_changed() {
        changed_map++;
    }

    public void changed() {
        changed( true );
    }

    public void changed( boolean redraw ) {
        changed++;

        if( redraw )
            repaint();
    }

    public void world_changed() {
        // FIXME: Remove this! CheeseMapData can get it's own events now
        CheeseMap.map.update_world();
        remap();
        map_changed();
    }

    public void keyTyped( KeyEvent e ) {	}
    public void keyReleased( KeyEvent e ) { }

    public void keyPressed( KeyEvent e ) {
        // Now control-arrows is move rooms, control-shift is +5, and just arrows is move map
        int x = 0,
            y = 0,
            z = 0;
        int delta = 1;

        if( (e.getModifiers() & InputEvent.SHIFT_MASK) != 0 ) {
            delta = 5;
        }

        if( e.getKeyCode() == KeyEvent.VK_UP ) {
            y -= delta;
        }else if( e.getKeyCode() == KeyEvent.VK_DOWN ) {
            y += delta;
        }else if( e.getKeyCode() == KeyEvent.VK_LEFT ) {
            x -= delta;
        }else if( e.getKeyCode() == KeyEvent.VK_RIGHT ) {
            x += delta;
        }else if( e.getKeyCode() == KeyEvent.VK_PAGE_UP ) {
            z++;
        }else if( e.getKeyCode() == KeyEvent.VK_PAGE_DOWN ) {
            z--;
        }

        if( x != 0 || y != 0 || z != 0 ) {
            if( (e.getModifiers() & InputEvent.CTRL_MASK) != 0 ) {
                // If we have selected items
                if( ! selected_rooms.isEmpty() ) {
                    Integer[] roomids = (Integer[])selected_rooms.keySet().toArray( new Integer[0] );

                    CheeseMap.move_rooms( roomids, x, y, z );
                }
            }else{
                move_origin( x, y, z );
            }
        }
    }

    public void set_tile_size( int width, int height ) {
        tile_width = width;
        tile_height = height;

        check_zoom();

        // Need to get new map data since our view changed
        remap();

        sendEvents();

        map_changed();
        changed();
    }

    // Perform merge for all items on screen
    public void merge_all() {
        cursor_saved = getCursor();
        setCursor( cursor_busy );

        /*
        TODO: busted, make it use our origin_x and width/height and perform the
              merge that way
        Object[] roomids = mapdata.values().toArray();
        for( int i = 0; i < roomids.length; i++ )
        	CheeseMap.send_tx( 1, "merge", ((Integer)roomids[ i ]).toString() );
        */

        setCursor( cursor_saved );
    }

    // Perform merge on all selected items
    // FIXME: Move this to CheeseMap class
    public void merge_selected() {
        cursor_saved = getCursor();
        setCursor( cursor_busy );

        Object[] roomids = selected_rooms.keySet().toArray();
        for( int i = 0; i < roomids.length; i++ )
            CheeseMap.send_tx( 1, "merge", ((Integer)roomids[ i ]).toString() );

        setCursor( cursor_saved );
    }

    public void setEditTool(String toolname) {
        if ( toolname.equals( toolnames[TOOL_SELECT] ) ) {
            setCursor( tool_cursors[TOOL_SELECT] );
            tool_active = TOOL_SELECT;
        }else if ( toolname.equals( toolnames[TOOL_MERGE] ) ) {
            setCursor( tool_cursors[TOOL_MERGE] );
            tool_active = TOOL_MERGE;
        }else if ( toolname.equals( toolnames[TOOL_MOVE] ) ) {
            setCursor( tool_cursors[TOOL_MOVE] );
            tool_active = TOOL_MOVE;
        }else if ( toolname.equals( toolnames[TOOL_DELETE] ) ) {
            setCursor( tool_cursors[TOOL_DELETE] );
            tool_active = TOOL_DELETE;
        }else if ( toolname.equals( toolnames[TOOL_POINT] ) ) {
            setCursor( tool_cursors[TOOL_POINT] );
            tool_saved = getEditTool();
            tool_active = TOOL_POINT;
        }
    }

    public String getEditTool() {
        return toolnames[tool_active];
    }

    private void save_exit(int from, String dir, int to) {
        CheeseLab.dmesg( 3, "Save exit: " + from + " - " + dir + " -> " + to );

        // FIXME: This has an unsolvable race condition.
        //        basically what we do is we fetch the room info and then send it back.
        //        This is a waste of bandwidth and falls apart if some1 changes
        //        room data in the meantime. Also waiting for a ping freezes the GUI.
        //        Creating an atomic exit change command in broker would be a lot
        //        better idea
        //        eg. CheeseMap.send_tx( 1, "save_exit", from + "," + dir + "," + to );

        // Make sure it gets in the cache
        CheeseRoomData room = CheeseMap.map.get_room( from, System.currentTimeMillis() );
        CheeseMap.send_tx(1, "ping", "set_edit_room");
        while ( !CheeseMap.checkping() );

        room.mtime = System.currentTimeMillis();
        if( dir.equalsIgnoreCase( "u" ) )
            room.exit_u = to;
        else if( dir.equalsIgnoreCase( "d" ) )
            room.exit_d = to;
        else if( dir.equalsIgnoreCase( "n" ) )
            room.exit_n = to;
        else if( dir.equalsIgnoreCase( "s" ) )
            room.exit_s = to;
        else if( dir.equalsIgnoreCase( "e" ) )
            room.exit_e = to;
        else if( dir.equalsIgnoreCase( "w" ) )
            room.exit_w = to;

        // Send save room command
        CheeseMap.save_room( room );
    }

    private void save_overlay(int from, String ovrl, boolean en) {
        Vector novrls = new Vector();

        CheeseLab.dmesg( 3, "Save overlay: " + from + " - " + ovrl + " -> " + en );

        // Make sure it gets in the cache
        CheeseRoomData room = CheeseMap.map.get_room( from, System.currentTimeMillis() );
        CheeseMap.send_tx(1, "ping", "set_edit_room");
        while ( !CheeseMap.checkping() );

        room.mtime = System.currentTimeMillis();
        // Copy all overlays except current one to vector
        for( int i = 0; i < room.overlays.length; i++ ) {
            String o = room.overlays[i];
            if( !o.equals("") && !o.equals( ovrl ) )
                novrls.add( o );
        }
        // Add current overlay if requested so
        if( en )
            novrls.add( ovrl );
        room.overlays = (String [])novrls.toArray( new String[0] );

        // Send save room command
        CheeseMap.save_room( room );
    }

    /**
     * Create custom cursor from a graphics file
     * @param name  name to be looked up. The filename takes form: ui/ui_cursor_name.png
     * @return      Cursor object with new cursor.
     */
    Cursor createCursor(String name) {
        // TODO: Scale cursor to system preferred size
        Toolkit tk = Toolkit.getDefaultToolkit();
        Image img = new ImageIcon( "ui/ui_cursor_" + name + ".png" ).getImage();
        return tk.createCustomCursor(img, new Point( 0,0 ), name);
    }

    /**
     * Implement this class to add callback to point tool. This tool can be used
     * to allow user to select just one room (to point a room).
     */
    abstract public class PointToolCallback {
        abstract void pointedTo(Point p);
    }

    /**
     * This class is used to implement Context Menu editor commands.
     */
    class CtxEditorListener implements ActionListener {
        String sdata;

        int roomid;

        CtxEditorListener(String sdata, int roomid) {
            this.sdata = sdata;
            this.roomid = roomid;
        }

        public void actionPerformed(ActionEvent e) {
            String cmd = e.getActionCommand();
            if(cmd.equals( "set_exit_none" )) {
                save_exit( roomid, sdata, 0 );
            }else if( cmd.equals( "set_exit_unexplored" ) ) {
                save_exit( roomid, sdata, -1 );
            }else if( cmd.equals( "set_exit_dt" ) ) {
                save_exit( roomid, sdata, -2 );
            }else if( cmd.equals( "set_exit_from_map" ) ) {
                point_callback = new PointToolCallback() {
                                     void pointedTo(Point p) {
                                         Point3d p3d = local_to_global( p.x, p.y );

                                         CheeseRoomData room = CheeseMap.map.get_room( p3d.x, p3d.y, p3d.z );
                                         if( room != null ) {
                                             save_exit( roomid, sdata, room.roomid );
                                         }
                                     }
                                 };
                setEditTool( "point_tool" );
            }else if( cmd.equals( "set_overlay" )) {
                JCheckBoxMenuItem src = (JCheckBoxMenuItem)e.getSource();
                save_overlay( roomid, sdata, src.isSelected() );
            }
        }
    }

    public Image get_image( String name, int width, int height ) {
        return( renderer.get_image( name, width, height ) );
    }

    public String tilename( CheeseRoomData r ) {
        String tilename = (String)tilenames.get( r.terrain );
        if( tilename == null )
            tilename = "unknown";
        return( tilename );
    }

    public Dimension get_tile_size() {
        Dimension ret = new Dimension( tile_width, tile_height );
        return( ret );
    }

    public boolean get_render_grid() {
        return( render_grid );
    }

    public boolean get_render_future() {
        return( render_future );
    }

    public boolean get_render_notes() {
        return( render_notes );
    }

    public int get_render_notes_length() {
        return( render_notes_length );
    }

    public Point3d get_origin() {
        Point3d ret = new Point3d();
        ret.x = origin_x;
        ret.y = origin_y;
        ret.z = origin_z;

        return( ret );
    }

    public Point3d get_map_origin() {
        Point3d ret = new Point3d();
        ret.x = map_origin_x;
        ret.y = map_origin_y;
        ret.z = map_origin_z;

        return( ret );
    }

    public void set_render_grid( boolean r ) {
        render_grid = r;
        map_changed();
        changed();
    }

    public void set_render_future( boolean r ) {
        render_future = r;
        map_changed();
        changed();
    }

    public void set_render_notes( boolean r ) {
        render_notes = r;
        map_changed();
        changed();
    }

    public String get_theme() {
        return( theme );
    }

    public boolean room_is_selected( int roomid ) {
        return( selected_rooms.get( new Integer( roomid ) ) != null );
    }

    public BufferedImage createCompatibleImage( int width, int height ) {
        return createCompatibleImage( width, height, Transparency.OPAQUE );
    }
    
    public BufferedImage createCompatibleImage( int width, int height, int transparency ) {
        if( screencache == null )
            newScreenCache();

        return screencache.createGraphics().getDeviceConfiguration().
               createCompatibleImage( width, height, transparency );
    }

    private void newScreenCache() {
        CheeseLab.dmesg( 2, "Creating new image for screen cache" );
        if (screencache != null) {
            screencache.flush();
            screencache = null;
        }         
        //screencache = (BufferedImage)createImage(getWidth(), getHeight());
        screencache = createVolatileImage( getWidth(), getHeight() );
        setImageAccelerationPriority( screencache, 1.0 );
    }
    
    private void newMapCache() {
        CheeseLab.dmesg( 2, "Creating new image for map cache" );
        if (mapcache != null) {
            mapcache.flush();
            mapcache = null;
        }         
        //mapcache = (BufferedImage)createImage(screencache.getWidth(), screencache.getHeight());
        mapcache = createVolatileImage( getWidth(), getHeight() );
        setImageAccelerationPriority( mapcache, 0.9 );
    }

    protected void updateMapOrigin() {
        map_origin_x = origin_x;
        map_origin_y = origin_y;
        map_origin_z = origin_z;
    }

    protected void sendEvents() {
        updateSelectionArea();
        
        for( int i = 0; i < listeners.size(); i++ ) {
            ((CheeseCanvasListener)listeners.elementAt( i )).viewChanged( this );
        }
    }

    public void addCanvasListener( CheeseCanvasListener l ) {
        listeners.add( l );
    }
    
    public void removeCanvasListener( CheeseCanvasListener l ) {
        listeners.remove( l );
    }

    public boolean roomInView( CheeseRoomData room ) {
        Point p = global_to_local( room.x, room.y, room.z );

        if( p.x < 0 || p.x > getWidth() ||
                p.y < 0 || p.y > getHeight() )
            return false;
        else
            return true;
    }

    public boolean get_follow_player() {
        return follow_player;
    }

    public void set_follow_player( boolean f ) {
        follow_player = f;
        follow_action.setSelected( f );

        // Get the playername and re-center on it if we can
        CheesePlayerData p = CheeseMap.map.get_player( follow_userid );
        if( p != null && follow_player ) {
            center_on( p.x, p.y, p.z );
        }
    }

    public String get_follow_userid() {
        return follow_userid;
    }

    public void set_follow_userid( String userid ) {
        follow_userid = userid;

        // Center on player too if needed
        set_follow_player( follow_player );
    }

    public void zoom_in() {
        resize_tiles( 5 );
        check_zoom();
    }

    public void zoom_out() {
        resize_tiles( -5 );
        check_zoom();
    }

    public void resize_tiles( int howmuch ) {
        Dimension tile_size = get_tile_size();
        set_tile_size( tile_size.width + howmuch,
                       tile_size.height + howmuch );
    }

    public void level_up() {
        set_follow_player( false );
        move_origin( 0, 0, 1 );
    }

    public void level_down() {
        set_follow_player( false );
        move_origin( 0, 0, -1 );
    }

    public void check_zoom() {
        Dimension tile_size = get_tile_size();

        if( tile_size.width <= 5)
            zoomout_action.setEnabled( false );
        else
            zoomout_action.setEnabled( true );
    }

    /* CheeseMapListener interface methods */

    public void roomChanged(CheeseRoomData newroom, CheeseRoomData oldroom) {
        // Check if room was selected & coordinates have been changed
        if( oldroom != null && room_is_selected( oldroom.roomid ) &&
            ( oldroom.x != newroom.x || oldroom.y != newroom.y ||
              oldroom.z != newroom.z ) ) {
            
            /* Take old room out of the selection area */
            Area a = (Area)selected_areas.get( new Integer( oldroom.z ) );
            a.subtract( new Area( new Rectangle( oldroom.x, oldroom.y, 1, 1 ) ) );
            
            /* Put it back with new coordinates */
            Rectangle r = new Rectangle( newroom.x, newroom.y, 1, 1 );
            a = (Area)selected_areas.get( new Integer( newroom.z ) );

            if( a == null ) {
                a = new Area( r );
                selected_areas.put( new Integer( newroom.z ), a );
            }else{
                a.add( new Area( r ) );
            }
            
            updateSelectionArea();
        }

        // Check if the change is in our visual area
        if( ! roomInView( newroom ) &&
                ( oldroom == null || ! roomInView( oldroom ) ) )
            return;

        world_changed();
        // Do not repaint on room event while refreshing (during sync cache)
        changed( ! refreshing );
    }

    public void roomDeleted(CheeseRoomData room ) {
        // Check if room was selected
        if( room != null && room_is_selected( room.roomid ) ) {
            Area a = (Area)selected_areas.get( new Integer( room.z ) );
            a.subtract( new Area( new Rectangle( room.x, room.y, 1, 1 ) ) );
            if( room.z == origin_z ) {
                updateSelectionArea();
            }
        }

        // Check if the change is in our visual area
        if( roomInView( room ) )
            world_changed();

        // Do not repaint on room event while refreshing (during sync cache)
        changed( ! refreshing );
    }

    public void playerChanged(CheesePlayerData p, CheesePlayerData op) {
        // FIXME: follow_player and follow_userid should be CheeseCanvas not CheeseMap properties
        if( p != null && follow_player && p.name.equals( follow_userid ) ) {
            CheeseLab.dmesg( 3, "Moving origin: " + p.x + "," + p.y + "," + p.z );
            center_on( p.x, p.y, p.z );
        }

        if( p.online || (op != null && op.online) )
            changed();
    }

    public void tileChanged(String fname) {
        // Invalidate map cache, this method currently does nothing but scream.
        // FIXME: implement this in BasicRenderer where it belongs
        //invalidate_cache();
    }

    public void refreshStarted() {
        refreshing = true;
    }

    public void refreshFinished() {
        refreshing = false;

        if( last_map < changed_map )
            changed();
    }

    public void onlineChanged(boolean online) {
        world_changed();
        changed();
    }


    public boolean get_render_hud() {
        return render_hud;
    }

    public void set_render_hud(boolean render_hud) {
        this.render_hud = render_hud;
        changed();
    }
    
    /* Another through bottom solution for supporting java 1.4 
     * FIXME: Remove when 1.4 support is dropped 
     */
    public void setImageAccelerationPriority( Image i, double pri ) {
        Class c = i.getClass();
        try {
            Method m = c.getMethod("setAccelerationPriority", new Class [] { float.class } );
            m.invoke( i, new Object [] { new Float( pri ) } );
        } catch (Exception e) { }
    }

    /**
     * Destroy this canvas beyond repair.
     */
    public void dispose() {
        // Deregister from map events
        CheeseMap.removeCheeseMapListener( this );

        // Kill renderer
        renderer.dispose();

        // Free data structures (to be collected by GC)
        listeners = null;
        selected_rooms = null;
        screencache = null;
        mapcache = null;
    }
}
