/*
 * Created on 23.10.2005
 */
package net.hoopajoo.cheeselab3;

import java.awt.*;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.image.*;
import java.util.*;
import java.util.Map.Entry;

/**
 * 
 * A modified CheeseCanvas to implement map view in CheeseMapNavigator window.
 * 
 * @author Antti Andreimann <Antti.Andreimann@mail.ee>
 *
 */
public class CheeseMapNavigatorCanvas extends CheeseCanvas implements CheeseCanvasListener {
    private BufferedImage sc = null;

    // Two points for tracking main map view position and size
    private Point mv_s = null;
    private Point mv_e = null;
    private int cols = 0, rows = 0;
    private int mv_cols, mv_rows;
    private Dimension scsize = new Dimension( 1, 1 );

    CheeseMapNavigatorCanvas() {
        super();

        /* Currently we use 1 pixel for navigator tile size, this might change
         *  in the future, if we decide to implement zoom in navigator window as well
         */
        set_tile_size( 1, 1 );

        set_render_future( false );
        set_render_notes( false );
        set_render_grid( false );

        // Hook us to receive canvas events
        CheeseMap.canvas.addCanvasListener( this );

        setPreferredSize( new Dimension(
                              CheeseMap.prop_get_int( "navi_w", 400 ),
                              CheeseMap.prop_get_int( "navi_h", 200 ) ) );
    }

    public void paintComponent(Graphics g) {

        if( CheeseMap.online == false ) {
            super.paintComponent( g );
            return;
        }

        // Need to repaint map
        if( ! refreshing && ( sc == null || last_map < changed_map ) ) {
            if( sc == null )
                sc = getGraphicsConfiguration().
                        createCompatibleImage( scsize.width, scsize.height );                

            /* FIXME: Currently all the remapping catchup logic has been omitted
             *        thus this code will probably break when sync_on_world = false
             *        however I consider running with sync_on_world off to be a legacy
             *        mode that should be removed sooner or later anyway
             */
            remap();

            updateMapOrigin();
            last_map = changed_map;

            // Bootstrap renderer
            renderer.prepare();

            // Render map
            Date rstart = new Date();
            renderer.draw_map( sc );
            Date rend = new Date();
            long rdiff = rend.getTime() - rstart.getTime();
            double secs = (double)rdiff / 1000;
            CheeseLab.dmesg( 2, "Navigator render time: " + rdiff + "ms (" + secs + " secs)" );
        }

        // Draw map
        if( sc != null )
            g.drawImage( sc, 0, 0, getWidth(), getHeight(), null );

        // Draw main map view box
        if( mv_s != null && mv_e != null ) {
            g.setColor( Color.yellow );
            Point s = osc_to_c( new Point( mv_s.x, mv_s.y ) );
            Point e = osc_to_c( new Point( mv_e.x, mv_e.y ) );
            g.drawRect( s.x , s.y, e.x - s.x, e.y - s.y );
        }
    }

    private Point c_to_osc( Point p ) {
        double scale_x = (double)scsize.width / getWidth();
        double scale_y = (double)scsize.height / getHeight();

        return new Point( (int)(p.x * scale_x), (int)(p.y * scale_y) );
    }

    private Point osc_to_c( Point p ) {
        double scale_x = (double)getWidth() / scsize.width;
        double scale_y = (double)getHeight() / scsize.height;

        return new Point( (int)(p.x * scale_x), (int)(p.y * scale_y) );
    }

    private void updateStats() {
        // Get map data statistics and set our origin accordingly
        int x = CheeseMap.map.stats.min_x;
        int y = CheeseMap.map.stats.min_y;

        // This means statistics are fucked up (this shouldn't happen)
        if( x >= Integer.MAX_VALUE || y >= Integer.MAX_VALUE ) {
            CheeseLab.dmesg( 2, "Unable to draw navigator (yet): Cache stats are missing." );
            return;
        }

        cols = CheeseMap.map.stats.max_x - x;
        rows = CheeseMap.map.stats.max_y - y;
        HashMap nrooms_z = CheeseMap.map.stats.nrooms_z;

        int z = 0;
        int nz = 0;

        Iterator i = nrooms_z.entrySet().iterator();
        while( i.hasNext() ) {
            Entry e = (Entry)i.next();
            Integer key = (Integer)e.getKey();
            Integer val = (Integer)e.getValue();
            if( nz < val.intValue() ) {
                nz = val.intValue();
                z = key.intValue();
            }
        }

        Dimension tmp = new Dimension( cols * get_tile_size().width,
                                       rows * get_tile_size().height );
        if( ! scsize.equals( tmp ) ) {
            sc = null;
            scsize = tmp;

            // Resize to correct aspect ratio
            setPreferredSize( new Dimension( getWidth(),
                                             (int)(scsize.height * (double)getWidth() / scsize.width ) ) );
            CheeseMap.navi.pack();

            // Recalculate main view box
            viewChanged( CheeseMap.canvas );
        }

        Point3d o = get_origin();

        if( o.x != x || o.y != y || o.z != z ) {
            // Set new origin
            set_origin( x, y, z );

            // Recalculate main view box
            viewChanged( CheeseMap.canvas );
        }
    }

    /* Mouse events */

    /* We don't want normal canvas mouse handling thus we override all
     * mouse callbacks and make most of them do absolutely nothing!
     */

    public void mouseReleased(MouseEvent e) { }
    public void mouseMoved(MouseEvent e) { }
    public void mouseEntered(MouseEvent e) { }
    public void mouseExited(MouseEvent e) { }
    public void mouseClicked(MouseEvent e) { }

    public void mousePressed(MouseEvent e) {
        Point l = c_to_osc( e.getPoint() );

        // Move main view
        Point3d g = local_to_global( l.x, l.y );
        g.x -= mv_cols / 2;
        g.y -= mv_rows / 2;

        CheeseMap.set_follow_player( false );
        CheeseMap.set_origin( g.x , g.y , g.z );
    }

    public void mouseDragged(MouseEvent e) {
        mousePressed( e );
    }
    
    // We don't support keys, mainly because we don't want user to
    // accidentally change the origin by pressing arrows or PgUp/PgDown
    public void keyPressed(KeyEvent e) { }

    public String getToolTipText( MouseEvent e ) {
        // on the second thought, tooltips in navigator window look rather silly.
        return null;

        /*
        // We have to do some coordinate translation magic to get correct tooltips
        // from the canvas.
        Point p = c_to_osc( e.getPoint() );
        e.translatePoint( -e.getX() + p.x, -e.getY() + p.y );

        return super.getToolTipText( e );
        */
    }

    public void playerChanged(CheesePlayerData p) {
        // We don't care about player events thus we
        // override this callback to save some repaints
    }

    public void viewChanged(CheeseCanvas c) {
        // Calculate map view box from main canvas
        Point3d p = c.get_origin();
        Dimension t = c.get_tile_size();
        mv_cols = c.getWidth() / t.width;
        mv_rows = c.getHeight() / t.height;

        mv_s = global_to_local( p.x, p.y, p.z );
        mv_e = global_to_local( p.x + mv_cols, p.y + mv_rows, p.z );

        repaint();
    }

    public void onlineChanged(boolean online) {
        super.onlineChanged( online );

        if( ! online )
            sc = null;
        else
            updateStats();
    }

    public void refreshFinished() {
        super.refreshFinished();

        updateStats();
    }
}
