/*
 *  
 *  Copyright (c) 2004 Steve Slaven, All Rights Reserved.
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Incanvas., 59 Temple Place, Suite 330, Boston,
 *  MA 02111-1307 USA
 *  
*/
package net.hoopajoo.cheeselab3;
import java.util.*;
import java.awt.*;
import java.awt.geom.Rectangle2D;
import java.awt.image.*;
import java.awt.event.*;
import javax.swing.*;
import net.hoopajoo.swingx.*;
import java.io.*;

// Main rendering class, inherit from this and override draw_map if you want to
// make a custom renderer
//
class CheeseRendererBasic extends CheeseRenderer {
    // Gridded overlays, TODO: make this a gui option
    boolean render_grid = true;

    /* Hacks */
    int tile_width, tile_height;
    int origin_x, origin_y, origin_z;
    private String theme;
    CheesePlayerData players[];

    // HashMap has faster get but slower put apparently, but it'll work
    /* According to Java docs, HashMap implementation is not thread safe.
     * Since we don't know what threads may call get_image and if they'll
     * do it concurrently or not, we have to get a synchronized instance.
     * We also make it static so all renderer instances share the same cache.
     */
    //private HashMap image_cache = new HashMap();
    private static Map image_cache = Collections.synchronizedMap(new HashMap());




    private Vector warplines = new Vector();

    public void prepare() {
        super.prepare();
        Dimension tile_size = canvas.get_tile_size();
        players = CheeseMap.map.get_players();
        tile_width = tile_size.width;
        tile_height = tile_size.height;
        Point3d origin = canvas.get_map_origin();
        origin_x = origin.x;
        origin_y = origin.y;
        origin_z = origin.z;
        theme = canvas.get_theme();
        render_grid = canvas.get_render_grid();
    }

    // This function draws just the map, not player tiles or HUD data
    // which may change without the map data moving
    public void draw_map( Image buffer ) {

        // We mark the sync now so that if there are changed while we
        // are rendering, we will get them
        int rend_x = origin_x,
            rend_y = origin_y,
            rend_z = origin_z;

        int cols = canvas.getWidth() / tile_width;
        int rows = canvas.getHeight() / tile_height;

        CheeseLab.dmesg( 3, "Size: " + rows + "," + cols );
        CheeseLab.dmesg( 3, "Origin: " + rend_x + "," + rend_y + "," + rend_z );

        // Figure our bounding box
        int start_x = rend_x;
        int start_y = rend_y;
        int start_z = rend_z;

        int end_x = start_x + cols;
        int end_y = start_y + rows;
        int end_z = rend_z + 1;

        warplines.clear();
        Graphics2D sg = (Graphics2D)buffer.getGraphics();

        // Make a slightly larger font
        sg.setFont( new Font( null, Font.BOLD, 12 ) );

        sg.setColor( new Color( 50, 50, 50 ) );
        sg.fillRect( 0, 0, canvas.getWidth(), canvas.getHeight() );

        Vector textnodes = new Vector();
        int x, y, z, local_x, local_y;
        for( z = 1; z >= 0; z-- ) {
            for( x = 0; x <= cols; x++ ) {
                local_x = x * tile_width;
                for( y = 0; y <= rows; y++ ) {
                    local_y = y * tile_height;
                    //System.out.println( "draw image: " + local_x + "," + local_y );
                    int global_x = start_x + x;
                    int global_y = start_y + y;
                    int global_z = start_z + z;

                    CheeseRoomData room = CheeseMap.map.get_room( global_x, global_y, global_z );
                    if( room != null ) {
                        String tilename = canvas.tilename( room );

                        sg.drawImage( get_image( "t_" + tilename, render_grid ), local_x, local_y, null );

                        String exittile;

                        exittile = get_exit_tile( room, room.exit_e, room.door_e, "e" );
                        if( exittile != null )
                            sg.drawImage( get_image( exittile ), local_x, local_y, null );
                        exittile = get_exit_tile( room, room.exit_w, room.door_w, "w" );
                        if( exittile != null )
                            sg.drawImage( get_image( exittile ), local_x, local_y, null );
                        exittile = get_exit_tile( room, room.exit_n, room.door_n, "n" );
                        if( exittile != null )
                            sg.drawImage( get_image( exittile ), local_x, local_y, null );
                        exittile = get_exit_tile( room, room.exit_s, room.door_s, "s" );
                        if( exittile != null )
                            sg.drawImage( get_image( exittile ), local_x, local_y, null );
                        exittile = get_exit_tile( room, room.exit_u, room.door_u, "u" );
                        if( exittile != null )
                            sg.drawImage( get_image( exittile ), local_x, local_y, null );
                        exittile = get_exit_tile( room, room.exit_d, room.door_d, "d" );
                        if( exittile != null )
                            sg.drawImage( get_image( exittile ), local_x, local_y, null );

                        if( global_z == rend_z ) {
                            boolean drew_an_overlay = false;
                            for(int i = 0; i < room.overlays.length; i++ ) {
                                if( ! room.overlays[ i ] .equals( "" ) ) {
                                    CheeseLab.dmesg( 3, "Overlay: " + room.overlays[ i ] );
                                    sg.drawImage( get_image( "overlay_" + room.overlays[ i ] ), local_x, local_y, null );
                                    drew_an_overlay = true;
                                }
                            }

                            // render notes text
                            if( room.notes != null && (! room.notes.equals( "" ) ) ) {
                                if( room.notes.length() < canvas.get_render_notes_length() && canvas.get_render_notes() ) {
                                    textnodes.add( room );
                                } else {
                                    if( ! drew_an_overlay ) {
                                        // Render a note tile if notes are here
                                        CheeseLab.dmesg( 3, "Doing default notes overlay" );
                                        sg.drawImage( get_image( "overlay_notes" ), local_x, local_y, null );
                                    }
                                }
                            }

                        } else {
                            sg.drawImage( get_image( "overlay_blacken" ), local_x, local_y, null );
                        }

                        // Selected
                        if( canvas.room_is_selected( room.roomid ) ) {
                            CheeseLab.dmesg( 3, "Drawing selected" );
                            sg.drawImage( get_image( "selected" ), local_x, local_y, null );
                        }

                        if( room.roomid == CheeseMap.edit_roomid )
                            sg.drawImage( get_image( "infobox" ), local_x, local_y, null );
                    }
                }
            }
        }

        // warp lines
        sg.setColor( new Color( 255, 0, 0 ) );
        for( int i = 0; i < warplines.size(); i++ ) {
            int[] points = (int[])warplines.get( i );
            Point pstart = canvas.global_to_local_reference( points[ 0 ], points[ 1 ], points[ 2 ],
                           rend_x, rend_y, rend_z );
            Point pend = canvas.global_to_local_reference( points[ 3 ], points[ 4 ], points[ 5 ],
                         rend_x, rend_y, rend_z );

            sg.drawLine( pstart.x + points [ 6 ], pstart.y + points[ 7 ],
                         pend.x + tile_width / 2, pend.y + tile_height / 2 );
        }

        // Notes
        for( int i = 0; i < textnodes.size(); i++ ) {
            CheeseRoomData room =(CheeseRoomData)textnodes.get( i );
            Point p = canvas.global_to_local_reference( room.x, room.y, room.z, rend_x, rend_y, rend_z );
            local_x = p.x;
            local_y = p.y;

            int offset_x = 4;
            int offset_y = tile_height - 4;

            if( room.x % 2 == 0 ) {
                offset_y = tile_height / 2;
            }

            sg.setColor( Color.BLACK );
            sg.drawString( room.notes,
                           local_x + offset_x, local_y + offset_y );
            sg.setColor( Color.LIGHT_GRAY );
            sg.drawString( room.notes,
                           local_x + offset_x + 1, local_y + offset_y + 1 );
        }

        sg.dispose();
    }


    public void draw_hud( Image buffer ) {
        if( players == null ) // Nothing to do (yet)
            return;
        
        //screencache = buffer;
        Graphics2D sg = (Graphics2D)buffer.getGraphics();

        // We mark the sync now so that if there are changed while we
        // are rendering, we will get them
        int rend_x = origin_x,
                     rend_y = origin_y,
                              rend_z = origin_z;

        // Player overlays
        CheeseLab.dmesg( 3, "Loading player map" );
        int hud_y = 5,
                    hud_x = canvas.getWidth() - 5;
        sg.setFont( new Font( null, Font.BOLD, 12 ) );
        sg.setColor( Color.WHITE );
        FontMetrics metrics = sg.getFontMetrics();
        for( int i = 0; i < players.length; i++ ) {
            if( players[ i ].online ) { // TODO: add selective showing later
                String username = players[ i ].name;
                if( username.equals( canvas.get_follow_userid() ) ) {
                    username = "YOU";
                }

                String player_info_string = username + ": " +
                                            players[ i ].hit_curr + "/" + players[ i ].hit_max + " hits " +
                                            players[ i ].mana_curr + "/" + players[ i ].mana_max + " mana " +
                                            players[ i ].move_curr + "/" + players[ i ].move_max + " moves";

                Rectangle2D bounds = metrics.getStringBounds( player_info_string, sg );
                sg.setColor( Color.BLACK );
                sg.drawString( player_info_string,
                               (int)(hud_x - bounds.getWidth()), (int)(hud_y + bounds.getHeight()) );
                sg.setColor( Color.WHITE );
                sg.drawString( player_info_string,
                               (int)(hud_x - bounds.getWidth()) + 1, (int)(hud_y + bounds.getHeight()) + 1 );

                hud_y += bounds.getHeight();
            }
        }

        sg.dispose();

    }

    public void draw_players( Image buffer ) {
        if( players == null ) // Nothing to do (yet)
            return;
        
        Graphics2D sg = (Graphics2D)buffer.getGraphics();

        // We mark the sync now so that if there are changed while we
        // are rendering, we will get them
        int rend_x = origin_x,
            rend_y = origin_y,
            rend_z = origin_z;

        // Player overlays
        CheeseLab.dmesg( 3, "Loading player map" );
        int hud_y = 5,
            hud_x = canvas.getWidth() - 5;
        sg.setFont( new Font( null, Font.BOLD, 12 ) );
        sg.setColor( Color.WHITE );
        FontMetrics metrics = sg.getFontMetrics();
        for( int i = 0; i < players.length; i++ ) {
            Point p = canvas.global_to_local_reference( players[ i ].x, players[ i ].y, players[ i ].z, rend_x, rend_y, rend_z );

            if( players[ i ].z == rend_z ) {
                if( p.x >= 0 && p.x <= canvas.getWidth() && p.y >= 0 && p.y <= canvas.getHeight() ) {
                    String playername = players[ i ].name;
                    sg.drawImage( get_image( "player_" + playername ), p.x, p.y, null );

                    if( players[ i ].roomid == 0 )
                        sg.drawImage( get_image( "player_lost" ), p.x, p.y, null );
                }
            }

            if( players[ i ].name.equals( canvas.get_follow_userid() ) && canvas.get_render_future() ) {
                // Draw where we are going to
                CheeseLab.dmesg( 3, "Drawing future placement" );
                Point last_point = canvas.global_to_local_reference( players[ i ].x, players[ i ].y, players[ i ].z,
                                   rend_x, rend_y, rend_z );
                last_point.x += (int)(tile_width / 2);
                last_point.y += (int)(tile_height / 2);
                sg.setColor( Color.YELLOW );
                for( int ii = 0; ii < players[ i ].future.length; ii++ ) {
                    if( ! players[ i ].future[ ii ].equals( "" ) ) {
                        String[] split = players[ i ].future[ ii ].split( "," );
                        if( ! split[ 1 ].equals( "0" ) ) {
                            Point next_point = canvas.global_to_local_reference(
                                                   Integer.parseInt( split[ 1 ] ),
                                                   Integer.parseInt( split[ 2 ] ),
                                                   Integer.parseInt( split[ 3 ] ),
                                                   rend_x, rend_y, rend_z );
                            next_point.x += (int)(tile_width / 2);
                            next_point.y += (int)(tile_height / 2);

                            sg.drawLine( last_point.x, last_point.y, next_point.x, next_point.y );
                            last_point = next_point;
                        }
                    }
                }
            }
        }

        sg.dispose();
    }

    private String get_exit_tile( CheeseRoomData room, int exit, String door, String dir ) {
        String tilename = null;

        if( exit == 0 ) {
            tilename = "overlay_noexit_" + dir;
        } else if( exit == -2 ) {
            tilename = "overlay_dt_" + dir;
        } else if( exit == -1 ) {
            tilename = "overlay_unexplored_" + dir;
        } else {
            if( ! door.equals( "" ) ) {
                if( door.startsWith( "-" ) ) {
                    tilename = "overlay_door_unlocked_" + dir;
                } else if( door.startsWith( "+" ) ) {
                    tilename = "overlay_door_locked_" + dir;
                } else {
                    tilename = "overlay_door_" + dir;
                }
            } else if( dir.equals( "u" ) || dir.equals( "d" ) ) {
                tilename = "overlay_exit_" + dir;
            }

            // Check for warping
            CheeseRoomData connect = CheeseMap.map.get_room( exit );

            // calc diff from this room to that room, if it's more than 1
            // than we need to render warp lines
            int roomdiff = 0;

            roomdiff += Math.abs( connect.x - room.x );
            roomdiff += Math.abs( connect.y - room.y );
            roomdiff += Math.abs( connect.z - room.z );

            if( roomdiff == 1 ) {
                // Make sure it's in the right direction
                if( dir.equals( "e" ) ) {
                    if( room.x - connect.x != -1 )
                        roomdiff = 0;
                } else if( dir.equals( "w" ) ) {
                    if( room.x - connect.x != 1 )
                        roomdiff = 0;
                } else if( dir.equals( "n" ) ) {
                    if( room.y - connect.y != 1 )
                        roomdiff = 0;
                } else if( dir.equals( "s" ) ) {
                    if( room.y - connect.y != -1 )
                        roomdiff = 0;
                } else if( dir.equals( "u" ) ) {
                    if( room.z - connect.z != -1 )
                        roomdiff = 0;
                } else if( dir.equals( "d" ) ) {
                    if( room.z - connect.z != 1 )
                        roomdiff = 0;
                }
            }

            if( roomdiff != 1 ) {
                //CheeseMap.dmesg( "Evaluating warp lines" );
                if( connect.z == room.z && room.z == origin_z ) {
                    // Draw warp lines
                    int offset_x = 0;
                    int offset_y = 0;
                    if( dir.equals( "e" ) ) {
                        offset_x = tile_width;
                        offset_y = tile_height / 2;
                    } else if( dir.equals( "w" ) ) {
                        offset_x = 0;
                        offset_y = tile_height / 2;
                    } else if( dir.equals( "n" ) ) {
                        offset_x = tile_width / 2;
                        offset_y = 0;
                    } else if( dir.equals( "s" ) ) {
                        offset_x = tile_width / 2;
                        offset_y = tile_height;
                    } else if( dir.equals( "u" ) ) {
                        offset_x = 3;
                        offset_y = 3;
                    } else if( dir.equals( "d" ) ) {
                        offset_x = tile_width - 3;
                        offset_y = tile_height - 3;
                    }

                    CheeseLab.dmesg( 3, "Adding warplines" );

                    int[] line = new int[ 8 ];
                    line[ 0 ] = room.x;
                    line[ 1 ] = room.y;
                    line[ 2 ] = room.z;

                    line[ 3 ] = connect.x;
                    line[ 4 ] = connect.y;
                    line[ 5 ] = connect.z;

                    line[ 6 ] = offset_x;
                    line[ 7 ] = offset_y;

                    warplines.add( line );
                } else {
                    // Change tile to "overlay_warp"
                    // because of diff z planes (connect is not visible)
                    tilename = "overlay_warp_" + dir;
                }
            }
        }

        return( tilename );
    }

    // Get image, no grid, current tilesize
    public Image get_image( String name ) {
        return( get_image( name, false ) );
    }

    // Do gridded at tile size
    public Image get_image( String name, boolean gridded ) {
        return( get_image( name, tile_width, tile_height, gridded ) );
    }

    // For other classes
    public Image get_image( String name, int width, int height ) {
        return( get_image( name, width, height, false ) );
    }

    public Image get_image( String name, int width, int height, boolean gridded ) {
        // Load an image
        // 1) Request and load from server first, then save to file
        // 2) Load from disk, search in local/theme, local, theme/

        // TODO: request from server? or no server-side tiles?

        String keyname = name + "-" + width + "-" + height + "-" + gridded;
        // If not in cache load it
        Image im = (Image)image_cache.get( keyname );
        if( im == null ) {
            CheeseLab.dmesg( 3, "Image not in cache, getting: " + keyname );
            Toolkit toolkit = Toolkit.getDefaultToolkit();

            // TODO: Try file in tiles/name.png then cacheprefix/tiles/name.png
            String tilenames[] = {
                                     "local/tiles/" + name + ".png",
                                     CheeseMap.cachedir + "/tiles/" + name + ".png",
                                     "tiles/" + theme + "/" + name + ".png",
                                     "tiles/default/" + name + ".png",
                                     "tiles/" + name + ".png"
                                 };

            CheeseLab.dmesg( 3, "Loading: " + name );
            for( int i = 0; i < tilenames.length; i++ ) {
                ImageIcon ii = new ImageIcon( tilenames[ i ] );
                if( ii.getIconWidth() != -1 ) {
                    CheeseLab.dmesg( 3, "Got match: " + tilenames[ i ] );
                    im = ii.getImage().getScaledInstance( width, height, Image.SCALE_SMOOTH );

                    // Force loading
                    while( ! canvas.prepareImage( im, -1, -1, canvas ) ) {
                        try {
                            // Be nice to system and let ImageFetcher threads do
                            // their work, we are not in a hurry here
                            Thread.sleep( 1 );
                        } catch( InterruptedException e ) {}
                    }


                    /* TODO: implement this here
                    // Alpha blending does not do acceleration
                    Image t = null;
                    if( renderer_cached_alpha ) {
                    t = sg.getDeviceConfiguration().createCompatibleImage( tile_width, tile_height, Transparency.TRANSLUCENT );
                    }else{
                    t = sg.getDeviceConfiguration().createCompatibleImage( tile_width, tile_height, Transparency.BITMASK );
                    }
                    */



                    // Alpha blending does not do acceleration, but I like it, and under non-win32 environments the accel doesn't work anyways :(
                    // so you all have to suffer!! *hop*
                    Image t = canvas.createCompatibleImage( width, height, Transparency.BITMASK );
                    //Image t = sg.getDeviceConfiguration().createCompatibleImage( tile_width, tile_height, Transparency.TRANSLUCENT );
                    Graphics gg = t.getGraphics();
                    gg.drawImage( im, 0, 0, null );

                    if( gridded )
                        gg.drawImage( get_image( "grid" ), 0, 0, null );

                    gg.dispose();

                    im = t;
                }
            }

            // Cache it in memory if we got one
            if( im != null ) {
                image_cache.put( keyname, im );
            }
        }

        // if we still don't have an image, return missing
        if( im == null )
            return( get_image( "missing", width, height, gridded ) );

        return( im );
    }
}
